import datetime
import re
from enum import Enum
from typing import Dict, Optional, Any, List, Union

import dateutil.parser

from imecilabt.gpulab.util.urn_util import URN, always_optional_urn

NVDATA_JOBDEFINITION = "jobDefinition"

RUN_DETAIL_SLAVE_HOSTNAME = 'slaveHostname'
RUN_DETAIL_SLAVE_DNSNAME = 'slaveDnsName'
RUN_DETAIL_SLAVE_INSTANCE_ID = 'slaveInstanceId'
RUN_DETAIL_SLAVE_INSTANCE_PID = 'slaveInstancePID'
RUN_DETAIL_GPU_IDS = 'gpuIds'
RUN_DETAIL_CPU_IDS = 'cpuIds'
RUN_DETAIL_MEMORY = 'memory'
RUN_DETAIL_WORKER_ID = 'workerId'
RUN_DETAIL_CLUSTER_ID = 'clusterId'
RUN_DETAIL_PORT_MAPPINGS = 'portMappings'
RUN_DETAIL_START_DATE = 'startDate'
RUN_DETAIL_END_DATE = 'endDate'
RUN_DETAIL_DEADLINE_DATE = 'deadlineDate'
RUN_DETAIL_NOTIFY_DATE = 'notifyDate'  # last time the user was notified about this job (about how long it is running)
RUN_DETAIL_SSH_USERNAME = 'sshUsername'
RUN_DETAIL_SSH_PROXY_USERNAME = 'sshProxyUsername'
RUN_DETAIL_SSH_PROXY_HOST = 'sshProxyHost'
RUN_DETAIL_SSH_PROXY_PORT = 'sshProxyPort'
RUN_DETAIL_SUMMARY_STATISTICS = 'summaryStatistics'
RUN_DETAIL_GPU_INFO = 'gpuFixedInfo'

DOCKER_IMAGE_USERPASS_PATTERN = re.compile('([a-zA-Z0-9_+-]*):([^@]*)@([^@]*)')

# Name components may contain lowercase letters, digits and separators.
# A separator is defined as a period, one or two underscores, or one or more dashes.
# A name component may not start or end with a separator.
# A tag name must be valid ASCII and may contain lowercase and uppercase letters,
# digits, underscores, periods and dashes. A tag name may not start with a period
# or a dash and may contain a maximum of 128 characters.
DOCKER_IMAGE_AND_TAG_PATTERN = re.compile('(.*):([a-zA-Z0-9_][a-zA-Z0-9_.-]*)')

class JobState(Enum):
    ONHOLD = 'On Hold'               # On hold, not planned to run at this time (not in queue)
    QUEUED = 'Queued'                # Available to run, waiting in queue
    ASSIGNED = 'Assigned'          # Assigned to a specific node by the scheduler, but not yet "picked up" by that node.
    CANCELLED = 'Cancelled'          # Cancelled during run (due to user request)
    TRANSFER = 'Transferring'        # Being sent to worker
    STARTING = 'Starting'            # Received by worker, setup in progress, not yet running
    RUNNING = 'Running'              # Running on worker
    FINISHED = 'Finished'            # Run completed
    DELETED = 'Deleted'              # Marked as deleted. This causes it to be ignored in "queue" view
    FAILED = 'Failed'                # Failure due to job definition problem, system problems, or other.


def _int_or_none(val: Optional[Any]) -> Optional[int]:
    return int(val) if val is not None else None


def urn_to_user_mini_id(urn: Union[str, URN]) -> Optional[str]:
    if not urn:
        return None
    u = always_optional_urn(urn)
    if u and u.authority and u.name:
        auth = u.authority
        if auth == 'wall2.ilabt.iminds.be':
            auth = 'w2'
        if auth == 'ilabt.imec.be':
            auth = 'ilabt'
        return '{name}@{auth}'.format(auth=auth, name=u.name)
    else:
        return None


def urn_to_auth(urn: Union[str, URN]) -> Optional[str]:
    if not urn:
        return None
    u = always_optional_urn(urn)
    if u and u.authority:
        return u.authority
    else:
        return None


def urn_to_name(urn: Union[str, URN]) -> Optional[str]:
    if not urn:
        return None
    u = always_optional_urn(urn)
    if u and u.name:
        return u.name
    else:
        return None


class Job:
    def __init__(self,
                 uuid: Optional[str],
                 name: str,
                 state: JobState,
                 nvdocker_data: Optional[Dict],
                 user_urn: str,
                 project_urn: str,
                 created: datetime.datetime,
                 state_updated: datetime.datetime,
                 queue_time: Optional[datetime.datetime],  # last time the job was put in the queue
                 gpulab_version: str,
                 run_details: Dict,
                 ssh_pubkeys: List[str],
                 emails_queue: List[str],
                 emails_run: List[str],
                 emails_end: List[str],
                 priority: int = 0,
                 user_email: Optional[str] = None,
                 ):
        # naive times not allowed
        assert queue_time is None or queue_time.tzinfo is not None
        assert created is None or created.tzinfo is not None
        assert state_updated is None or state_updated.tzinfo is not None

        self.uuid = uuid
        self.id = uuid   #id and uuid are synonyms for Job
        self.name = name
        self.state = state
        self.nvdocker_data = nvdocker_data  # dict with "jobDefinition" element (which contains the interesting dict)
        self.user_urn = user_urn
        self.project_urn = project_urn
        self.user_email = user_email
        self.created = created
        self.state_updated = state_updated
        self.queue_time = queue_time
        self.gpulab_version = gpulab_version
        self.priority = priority
        # self.run_details : Dict = run_details if run_details is not None else {}
        # self.ssh_pubkeys : List[str] = ssh_pubkeys if ssh_pubkeys is not None else []
        # self.emails_queue : List[str] = emails_queue if emails_queue is not None else []
        # self.emails_run : List[str] = emails_run if emails_run is not None else []
        # self.emails_end : List[str] = emails_end if emails_end is not None else []
        # Annotations like above not supported before 3.6
        self.run_details = run_details if run_details is not None else {}
        self.ssh_pubkeys = ssh_pubkeys if ssh_pubkeys is not None else []
        self.emails_queue = emails_queue if emails_queue is not None else []
        self.emails_run = emails_run if emails_run is not None else []
        self.emails_end = emails_end if emails_end is not None else []

    def _get_job_def_attr(self, attr_name: str) -> Any:
        if self.nvdocker_data is None \
                or not NVDATA_JOBDEFINITION in self.nvdocker_data \
                or not attr_name in self.nvdocker_data[NVDATA_JOBDEFINITION]:
            return None
        return self.nvdocker_data[NVDATA_JOBDEFINITION][attr_name]

    def _get_job_run_detail_attr(self, attr_name: str) -> Optional[Any]:
        if self.run_details is None \
                or not attr_name in self.run_details:
            return None
        return self.run_details[attr_name]

    def _set_job_run_detail_attr(self, attr_name: str, new_value: Optional[Any]) -> None:
        if self.run_details is None:
            return
        if new_value is None:
            if attr_name in self.run_details:
                del self.run_details[attr_name]
        else:
            self.run_details[attr_name] = new_value

    def _get_job_def_attr_attr(self, attr_level1_name: str, attr_level2_name: str) -> Any:
        if self.nvdocker_data is None \
                or not NVDATA_JOBDEFINITION in self.nvdocker_data \
                or not attr_level1_name in self.nvdocker_data[NVDATA_JOBDEFINITION]:
            return None
        level1_attr = self.nvdocker_data[NVDATA_JOBDEFINITION][attr_level1_name]
        if level1_attr is None \
                or not attr_level2_name in level1_attr:
            return None
        return level1_attr[attr_level2_name]

    # nvdata jobDefinition getters

    def get_name(self) -> str:
        res = self._get_job_def_attr('name')
        assert self.name == res, "{} != {}".format(self.name, res)
        return res

    def get_description(self) -> str:
        return self._get_job_def_attr('description')

    def get_user_email(self) -> Optional[str]:
        if self.user_email:
            return self.user_email
        else:
            # fallback for wall2  TODO remove this legacy
            if self.user_urn and 'all2.ilabt.iminds.be' in self.user_urn:
                return '{}@wall2.ilabt.iminds.be'.format(self.get_username())
            return None

    def get_username(self) -> str:
        if self.user_urn and self.user_urn.startswith('urn:'):
            return URN(urn=self.user_urn).name
        else:
            return None

    def get_user_mini_id(self) -> str:
        if self.user_urn and self.user_urn.startswith('urn:'):
            return urn_to_user_mini_id(self.user_urn)
        else:
            return None

    def get_projectname(self) -> str:
        if self.project_urn and self.project_urn.startswith('urn:'):
            return URN(urn=self.project_urn).name
        else:
            return None

    def get_requested_cluster_ids(self) -> List[int]:
        """:return a list of cluster IDs that the user allows. Can be an empty list, but never None. An empty list means: accept any cluster."""
        res = self._get_job_def_attr('clusterId')
        if res is None:
            return []
        elif isinstance(res, int):
            return [res]
        elif isinstance(res, str):
            return [int(res)]
        else:
            assert isinstance(res, (list, tuple))
            assert not isinstance(res, str)
            return list(map(int, res))

    def get_docker_image(self) -> Optional[str]:
        return self._get_job_def_attr('dockerImage')

    def get_docker_image_nopass(self) -> Optional[str]:
        if self._get_job_def_attr('dockerImage'):
            res = self._get_job_def_attr('dockerImage')
            match = DOCKER_IMAGE_USERPASS_PATTERN.match(res)
            if match:
                return "{}:XXXX@{}".format(match.group(1), match.group(3))
            else:
                return res
        else:
            return None

    def get_job_type(self) -> str:
        return self._get_job_def_attr('jobType')

    def get_command(self) -> str:
        return self._get_job_def_attr('command')

    def get_resources_gpus(self) -> int:
        res = _int_or_none(self._get_job_def_attr_attr('resources', 'gpus'))
        return res

    def get_resources_system_memory(self) -> int:
        res = _int_or_none(self._get_job_def_attr_attr('resources', 'systemMemory'))
        return res

    def get_resources_cpu_cores(self) -> int:
        res = _int_or_none(self._get_job_def_attr_attr('resources', 'cpuCores'))
        return res

    def get_resources_gpu_model(self) -> Union[List[str], str]:
        res = self._get_job_def_attr_attr('resources', 'gpuModel')
        return res

    def get_resources_gpu_model_likepatterns(self) -> List[str]:
        res = self._get_job_def_attr_attr('resources', 'gpuModel')
        if res is None:
            return []
        if isinstance(res, str):
            return ['%{}%'.format(res)]
        if hasattr(res, "__getitem__") or hasattr(res, "__iter__"): # Some type of iterable/List
            return list(map(lambda r: '%{}%'.format(r), res))
        raise ValueError('unsupported resources.gpuModel type: {}'.format(type(res)))

    def get_resources_min_cuda_version(self) -> int:
        res = _int_or_none(self._get_job_def_attr_attr('resources', 'minCudaVersion'))
        return res

    def get_job_data_locations(self) -> List:
        return self._get_job_def_attr('jobDataLocations')

    def get_requested_port_mappings(self) -> List:
        return self._get_job_def_attr('portMappings')

    # run_details getters

    def get_worker_hostname(self) -> str:
        """alias for get_slave_hostname"""
        return self.get_slave_hostname()

    def get_slave_hostname(self) -> str:
        return self._get_job_run_detail_attr(RUN_DETAIL_SLAVE_HOSTNAME)

    def get_worker_dnsname(self) -> str:
        """alias for get_slave_dnsname"""
        return self.get_slave_dnsname()

    def get_slave_dnsname(self) -> str:
        return self._get_job_run_detail_attr(RUN_DETAIL_SLAVE_DNSNAME)

    def get_slave_instance_id(self) -> str:
        return self._get_job_run_detail_attr(RUN_DETAIL_SLAVE_INSTANCE_ID)

    def get_slave_instance_pid(self) -> int:
        res = self._get_job_run_detail_attr(RUN_DETAIL_SLAVE_INSTANCE_PID)
        return int(res) if res is not None else None

    def get_gpu_ids(self) -> List[str]:
        return self._get_job_run_detail_attr(RUN_DETAIL_GPU_IDS)

    def get_cpu_ids(self) -> List[str]:
        return self._get_job_run_detail_attr(RUN_DETAIL_CPU_IDS)

    def get_allocated_memory(self) -> List[str]:
        return self._get_job_run_detail_attr(RUN_DETAIL_MEMORY)

    def get_worker_id(self) -> int:
        return self._get_job_run_detail_attr(RUN_DETAIL_WORKER_ID)

    def get_cluster_id(self) -> int:
        return self._get_job_run_detail_attr(RUN_DETAIL_CLUSTER_ID)

    def get_port_mappings(self) -> List:
        return self._get_job_run_detail_attr(RUN_DETAIL_PORT_MAPPINGS)

    def get_ssh_username(self) -> Optional[str]:
        return self._get_job_run_detail_attr(RUN_DETAIL_SSH_USERNAME)

    def get_ssh_proxy_username(self) -> Optional[str]:
        return self._get_job_run_detail_attr(RUN_DETAIL_SSH_PROXY_USERNAME)

    def get_ssh_proxy_host(self) -> Optional[str]:
        return self._get_job_run_detail_attr(RUN_DETAIL_SSH_PROXY_HOST)

    def get_ssh_proxy_port(self) -> Optional[int]:
        return self._get_job_run_detail_attr(RUN_DETAIL_SSH_PROXY_PORT)

    def get_start(self) -> Optional[datetime.datetime]:
        s = self._get_job_run_detail_attr(RUN_DETAIL_START_DATE)
        if s is None:
            return None
        else:
            return dateutil.parser.parse(s)

    def get_duration(self) -> Optional[datetime.timedelta]:
        start = self._get_job_run_detail_attr(RUN_DETAIL_START_DATE)
        end = self._get_job_run_detail_attr(RUN_DETAIL_END_DATE)
        if start is None:
            return None
        else:
            start = dateutil.parser.parse(start)
        if end is None:
            end = datetime.datetime.now(datetime.timezone.utc)
        else:
            end = dateutil.parser.parse(end)
        assert end.tzinfo is not None, 'end is naive "{}"'.format(self._get_job_run_detail_attr(RUN_DETAIL_END_DATE))
        assert start.tzinfo is not None, 'start is naive "{}"'.format(self._get_job_run_detail_attr(RUN_DETAIL_START_DATE))
        return end - start

    def get_queue_duration(self) -> Optional[datetime.timedelta]:
        start = self.queue_time
        end = self._get_job_run_detail_attr(RUN_DETAIL_START_DATE)
        if start is None:
            return None
        if end is None:
            end = datetime.datetime.now(datetime.timezone.utc)
        else:
            end = dateutil.parser.parse(end)
        assert end.tzinfo is not None, 'end is naive "{}"'.format(self._get_job_run_detail_attr(RUN_DETAIL_START_DATE))
        assert start.tzinfo is not None, 'start is naive "{}"'.format(self.queue_time)
        return end - start

    def get_end(self) -> Optional[datetime.datetime]:
        s = self._get_job_run_detail_attr(RUN_DETAIL_END_DATE)
        if s is None:
            return None
        else:
            return dateutil.parser.parse(s)

    def get_last_notify_date(self) -> Optional[datetime.datetime]:
        s = self._get_job_run_detail_attr(RUN_DETAIL_NOTIFY_DATE)
        if s is None:
            return None
        else:
            return dateutil.parser.parse(s)

    def set_last_notify_date(self, notify_date: Optional[datetime.datetime]) -> None:
        notify_date_str = notify_date.isoformat() if notify_date is not None else None
        self._set_job_run_detail_attr(RUN_DETAIL_NOTIFY_DATE, notify_date_str)

    def get_deadline(self) -> Optional[datetime.datetime]:
        s = self._get_job_run_detail_attr(RUN_DETAIL_DEADLINE_DATE)
        if s is None:
            return None
        else:
            return dateutil.parser.parse(s)

    # example:
    # "jobDefinition": {
    #      "name": "Iterativenet",
    #      "description": "hello world",
    #      "clusterId": 129,
    #      "dockerImage": "gpulab.ilabt.imec.be:5000/sample:nvidia-smi",
    #      "jobType": "BATCH",
    #      "command": "",
    #      "resources": {
    #          "gpus": 1,
    #          "systemMemory": 32768,
    #          "cpuCores": 2
    #      },
    #      "jobDataLocations": [ ],
    #      "portMappings": [ ]
    #  }

    def __repr__(self):
        return 'GpuLabJob %s: %s' % (self.name, self.uuid)

    def __eq__(self, rhs):
        return (
                self.uuid == rhs.uuid and
                type(self) == type(rhs))

    def full_equal(self, rhs):
        return (
                self.uuid == rhs.uuid and
                self.name == rhs.name and
                self.state == rhs.state and
                self.nvdocker_data == rhs.nvdocker_data and
                self.user_urn == rhs.user_urn and
                self.project_urn == rhs.project_urn and
                self.user_email == rhs.user_email and
                self.created == rhs.created and
                self.state_updated == rhs.state_updated and
                self.gpulab_version == rhs.gpulab_version and
                self.run_details == rhs.run_details and
                self.ssh_pubkeys == rhs.ssh_pubkeys and
                type(self) == type(rhs))

    def sanitized_copy(self, logged_in=True) -> 'Job':
        """copy with removed confidential data
        :param logged_in: sanitized copy for logged in user, or for anonymous user?
        """
        run_details_sancopy = { }
        for public_detail in (RUN_DETAIL_CLUSTER_ID, RUN_DETAIL_START_DATE,
                              RUN_DETAIL_END_DATE, RUN_DETAIL_DEADLINE_DATE):
            if public_detail in self.run_details:
                run_details_sancopy[public_detail] = self.run_details[public_detail]
        if logged_in:
            for public_detail in (RUN_DETAIL_SLAVE_HOSTNAME, RUN_DETAIL_SLAVE_DNSNAME,
                                  RUN_DETAIL_SLAVE_INSTANCE_ID, RUN_DETAIL_SLAVE_INSTANCE_PID,
                                  RUN_DETAIL_NOTIFY_DATE, RUN_DETAIL_SUMMARY_STATISTICS):
                if public_detail in self.run_details:
                    run_details_sancopy[public_detail] = self.run_details[public_detail]

        nvdocker_data_sancopy = { }
        nvdocker_data_sancopy[NVDATA_JOBDEFINITION] = { }
        if self.get_name():
            nvdocker_data_sancopy[NVDATA_JOBDEFINITION]['name'] = self.get_name()
        if logged_in and self.get_description():
            nvdocker_data_sancopy[NVDATA_JOBDEFINITION]['description'] = self.get_description()
        if self.get_requested_cluster_ids():
            nvdocker_data_sancopy[NVDATA_JOBDEFINITION]['clusterId'] = self.get_requested_cluster_ids()
        if logged_in and self._get_job_def_attr('resources'):
            nvdocker_data_sancopy[NVDATA_JOBDEFINITION]['resources'] = self._get_job_def_attr('resources')

        copy = Job(
            uuid=self.uuid,
            name=self.name,
            state=self.state,
            nvdocker_data=nvdocker_data_sancopy, # self.nvdocker_data,
            user_urn=self.user_urn,
            project_urn=self.project_urn,
            user_email='known by server' if self.user_email else None,  # email is not exposed in sanitized copies
            created=self.created,
            state_updated=self.state_updated,
            queue_time=self.queue_time,
            gpulab_version=self.gpulab_version,
            run_details=run_details_sancopy,
            ssh_pubkeys=[], # no need for others to see pubkeys, even though it can't really harm
            emails_queue=[],
            emails_run=[],
            emails_end=[],
            priority=0
        )
        return copy

    def to_dict(self) -> dict:
        res = dict()
        res['uuid'] = self.uuid
        res['name'] = self.name
        res['state'] = self.state.name if self.state is not None else None
        res['nvdocker_data'] = self.nvdocker_data
        res['user_urn'] = self.user_urn
        res['userurn_auth'] = urn_to_auth(self.user_urn)
        res['userurn_name'] = urn_to_name(self.user_urn)
        res['user_mini_id'] = urn_to_user_mini_id(self.user_urn)
        res['username'] = res['user_mini_id']  # to have some backward compatibility
        res['project_name'] = self.get_projectname()
        res['project_urn'] = self.project_urn
        res['user_email'] = self.user_email
        res['created'] = self.created.isoformat() if self.created is not None else None
        res['state_updated'] = self.state_updated.isoformat() if self.state_updated is not None else None
        res['queue_time'] = self.queue_time.isoformat() if self.queue_time is not None else None
        res['gpulab_version'] = self.gpulab_version
        res['run_details'] = self.run_details
        res['ssh_pubkeys'] = self.ssh_pubkeys
        res['emails_queue'] = self.emails_queue
        res['emails_run'] = self.emails_run
        res['emails_end'] = self.emails_end
        if self.priority and self.priority != 0:
            res['priority'] = self.priority
        return res

    @classmethod
    def from_dict(cls, d: dict):
        return cls(
            uuid=d['uuid'],
            name=d['name'],
            state=JobState[d['state']] if 'state'in d and d['state'] else None,
            nvdocker_data=d['nvdocker_data'],
            user_urn=d['user_urn'],
            project_urn=d['project_urn'],
            user_email=d['user_email'] if 'user_email' in d and d['user_email'] else None,
            created=dateutil.parser.parse(d['created']) if 'created' in d and d['created'] else None,
            state_updated=dateutil.parser.parse(d['state_updated']) if 'state_updated' in d and d['state_updated'] else None,
            queue_time=dateutil.parser.parse(d['queue_time']) if 'queue_time' in d and d['queue_time'] else None,
            gpulab_version=d['gpulab_version'],
            run_details=d['run_details'],
            ssh_pubkeys=d['ssh_pubkeys'] if 'ssh_pubkeys' in d else None,
            emails_queue=d['emails_queue'] if 'emails_queue' in d else None,
            emails_run=d['emails_run'] if 'emails_run' in d else None,
            emails_end=d['emails_end'] if 'emails_end' in d else None,
            priority=d['priority'] if 'priority' in d else None
        )

# >>> import json
# >>> class ComplexEncoder(json.JSONEncoder):
# ...     def default(self, obj):
# ...         if isinstance(obj, complex):
# ...             return [obj.real, obj.imag]
# ...         # Let the base class default method raise the TypeError
# ...         return json.JSONEncoder.default(self, obj)
# ...
# >>> json.dumps(2 + 1j, cls=ComplexEncoder)

# Note: Job ID is a 36 char UUID (thus including the hyphens)


def is_partial_job_id(job_id: str):
    if re.search("[^0-9a-fA-F-]", job_id):
        return False
    return len(job_id) < 36


def is_full_job_id(job_id: str):
    if is_bad_job_id(job_id):
        return False
    return len(job_id) == 36


def is_bad_job_id(job_id: str):
    """Identifies job_id for which there is 100% certain a problem (too short, wrong chars, too long)"""
    l = len(job_id)
    if l != 36:
        return True
    if re.search("[^0-9a-fA-F-]", job_id):
        return True
    # TODO: also check UUID pattern here (8-4-4-4-12)
    return False

