from typing import Optional, List, Union

from imecilabt.gpulab.model.job import JobState, _int_or_none


class JobFilter:
    def __init__(self,
                 version: Optional[str],
                 cluster_id: Optional[Union[int, str]] = None,
                 empty_cluster_id: bool = True,  # If cluster_id is set, also match jobs that do not have clusterId?  If cluster_id not set, this is implicitly always True!
                 cpu_sharing_jobs: bool = True,  # Also match jobs that want to share a CPU? (jobs with 0 CPU requested)
                 max_system_memory_mb: Optional[int] = None,
                 max_gpu_count: Optional[int] = None,
                 max_cpu_count: Optional[int] = None,
                 allowed_states: Optional[List[JobState]] = None, # [JobState.QUEUED] is a common value  [] and None mean don't filter
                 user_urn: Optional[str] = None,
                 other_user_running: bool = False,
                 project_urn: Optional[str] = None,
                 max_duration_s: Optional[int] = None,
                 slave_cuda_version: Optional[int] = None,  # Only major cuda version should be passed
                 slave_gpu_models: Optional[List[str]] = None,
                 user_name: Optional[str] = None,
                 project_name: Optional[str] = None,):
        self.version = version  #dev or stable
        self.allowed_states = allowed_states  # if allowed_states else None
        self.cluster_id: Optional[int] = int(cluster_id) if cluster_id is not None else None
        self.empty_cluster_id = empty_cluster_id
        self.cpu_sharing_jobs = cpu_sharing_jobs
        assert cluster_id is not None or empty_cluster_id  # empty_cluster_id may not be False when cluster_id == None
        self.max_gpu_count = _int_or_none(max_gpu_count)
        self.max_cpu_count = _int_or_none(max_cpu_count)
        self.max_system_memory_mb = _int_or_none(max_system_memory_mb)
        self.user_urn = user_urn
        self.other_user_running = other_user_running
        self.project_urn = project_urn
        self.max_duration_s = _int_or_none(max_duration_s)
        self.slave_cuda_version = slave_cuda_version
        self.slave_gpu_models = [] if slave_gpu_models is None else list(slave_gpu_models)
        self.user_name = user_name
        self.project_name = project_name

    @classmethod
    def forSlave(cls,
                 version: str,
                 cluster_id: int,
                 empty_cluster_id: bool,
                 cpu_sharing_jobs: bool,
                 max_system_memory_mb: int,
                 max_gpu_count: Optional[int] = 1,
                 max_cpu_count: Optional[int] = 1,
                 slave_cuda_version: Optional[int] = None,
                 slave_gpu_models: Optional[List[str]] = None) -> 'JobFilter':
        """typical JobFilter used by slave to search a Job to execute"""
        return cls(
            version=version,
            cluster_id=cluster_id,
            empty_cluster_id=empty_cluster_id,
            cpu_sharing_jobs=cpu_sharing_jobs,
            max_system_memory_mb=max_system_memory_mb,
            max_gpu_count=max_gpu_count,
            max_cpu_count=max_cpu_count,
            allowed_states=[JobState.QUEUED],
            slave_cuda_version=slave_cuda_version,
            slave_gpu_models=slave_gpu_models
        )

    @classmethod
    def forPending(cls,
                   version: str) -> 'JobFilter':
        """typical JobFilter used to find all Pending Jobs"""
        return cls(
            version=version,
            allowed_states=[JobState.QUEUED]
        )

    @classmethod
    def noFilter(cls):
        return cls(
            version=None,
            allowed_states=[]
        )

    def __str__(self):
        return '%s(%s)' % (
            type(self).__name__,
            ', '.join('%s=%s' % item for item in vars(self).items())
        )

    def to_dict(self) -> dict:
        res = dict()
        res['version'] = self.version
        res['cluster_id'] = self.cluster_id
        res['empty_cluster_id'] = self.empty_cluster_id
        res['cpu_sharing_jobs'] = self.cpu_sharing_jobs
        res['max_system_memory_mb'] = self.max_system_memory_mb
        res['max_gpu_count'] = self.max_gpu_count
        res['max_cpu_count'] = self.max_cpu_count
        res['allowed_states'] = list(map(lambda state: state.name, self.allowed_states)) if self.allowed_states is not None else None
        res['user_urn'] = self.user_urn
        res['other_user_running'] = self.other_user_running
        res['project_urn'] = self.project_urn
        res['max_duration_s'] = self.max_duration_s
        res['slave_cuda_version'] = self.slave_cuda_version
        res['slave_gpu_models'] = self.slave_gpu_models if self.slave_gpu_models is not None else None
        res['user_name'] = self.user_name
        res['project_name'] = self.project_name
        return res

    @classmethod
    def from_dict(cls, d: dict):
        return cls(
            version=d.get('version'),
            cluster_id=d.get('cluster_id'),
            empty_cluster_id=d.get('empty_cluster_id'),
            cpu_sharing_jobs=d.get('cpu_sharing_jobs'),
            max_system_memory_mb=d.get('max_system_memory_mb'),
            max_gpu_count=d.get('max_gpu_count'),
            max_cpu_count=d.get('max_cpu_count'),
            allowed_states=list(map(lambda state: JobState[state], d.get('allowed_states'))) if d.get('allowed_states') else None,
            user_urn=d.get('user_urn'),
            other_user_running=d.get('other_user_running'),
            project_urn=d.get('project_urn'),
            max_duration_s=d.get('max_duration_s'),
            slave_cuda_version=d.get('slave_cuda_version'),
            slave_gpu_models=d.get('slave_gpu_models'),
            user_name=d.get('user_name'),
            project_name=d.get('project_name'),
        )
