import copy
from collections import deque
from typing import Optional, Any

from click import ParamType


class KeyValueParamType(ParamType):
    name = 'keyvalue'

    def convert(self, value, param, ctx):
        split = value.split('=')

        if len(split) != 2:
            self.fail("%s is not a valid key=value" % value, param, ctx)

        return (split[0], split[1])


class VolumeMappingParamType(ParamType):
    name = 'volumemapping'

    def convert(self, value, param, ctx):
        split = value.split(':')

        if len(split) > 3:
            self.fail("%s is not a valid volumemapping: expected 3 parts max, but got %d" % (value, len(split)),
                      param, ctx)

        host_src = container_dest = split[0]
        if len(split) > 1:
            container_dest = split[1]

        options = "rw"
        if len(split) > 2:
            options = split[2]

        return {"host_src": host_src, "container_dest": container_dest, "options": options}


def any_to_opt_bool(value: Any, default: Optional[bool] = None) -> Optional[bool]:
    if value is None:
        return default
    if value == 1:
        return True
    if value == 0:
        return False
    if str(value).lower() in ['true', 't', '1', 'yes']:
        return True
    if str(value).lower() in ['false', 'f', '0', 'no']:
        return False
    return default


def any_to_bool(value: Any, default: Optional[bool] = None) -> bool:
    res = any_to_opt_bool(value, default)
    if res is not None:
        return res
    if default is None:
        raise 'Could not convert "{}" to boolean value'.format(value)
    return default


def deep_update_dict(target: dict, extra: dict) -> dict:
    """
    Deep update by adding all keys (nested) of extra to target.
    Nested arrays are copied from extra to target with copy.deepcopy(), so no deep_update is done on them!

    Notes:
       - this deeply modifies target, so make sure you first copy.deepcopy it if needed!
       - you can't remove keys from target with this, only add them
       - type conflicts between values of target and extra are handled by extra overriding target

    :param target: the dict that should be updated deeply
    :param extra: extra keys to be added (nested)
    :return: target
    """

    # deque to avoid recursion
    jobs = deque()
    jobs.append( (target, extra) )

    while len(jobs) > 0:
        job_target, job_extra = jobs.popleft()

        for key,extra_value in job_extra.items():
            if isinstance(extra_value, dict):
                if key in job_target and isinstance(job_target[key], dict):
                    jobs.append( (job_target[key], extra_value) )
                else:
                    job_target[key] = copy.deepcopy(extra_value)
            else:
                job_target[key] = extra_value

    return target


# Not best idea: need to differentiate between key not present and bool parse failure
# def any_dictvalue_to_opt_bool(d: dict, key: Any, default: Optional[bool] = None) -> Optional[bool]:
#     value = d.get(key, default=None)
#     return any_to_opt_bool(value, default)
#
#
# def any_dictvalue_to_bool(d: dict, key: Any, default: Optional[bool] = None) -> bool:
#     res = any_dictvalue_to_opt_bool(d, key, default)
#     if res is not None:
#         return res
#     if default is None:
#         raise 'Could not convert "{}" to boolean value'.format(value)
#     return default

